/* PROGRAM: Talk the (parallel) talk, walk the (random) walk
 * AUTHORS: Aaron Weeden (Shodor) and Skylar Thompson (University of Washington)
 * DATE: November 2013
 */

#include <errno.h>
#include <getopt.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

typedef struct {
    bool isEmpty;
    bool willBeEmpty;
} cell_t;

typedef struct {
    int cellIndex;
} agent_t;

int main(int argc, char **argv) {
    int rowCount = 22;
    int columnCount = 22;
    int agentCount = 50;
    int timeStepCount = 100;
    int randomSeed;
    int cellCount;
    cell_t *cellList;
    agent_t *agentList;
    int cellIndex;
    int newCellIndex;
    int agentIndex;
    int timeStepIndex;
    int direction;
    int originalDirection;
    int opt;

    randomSeed = time(NULL) * getpid();

    /* Get user input */
    while((opt = getopt(argc,argv,"t:a:r:c:s:")) != -1) {
        switch(opt) {
            case 't':
                errno = 0;
                timeStepCount = strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"timeStepCount conv failed\n");
                    abort();
                }
                break;
            case 'a':
                errno = 0;
                agentCount = strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"agentCount conv failed\n");
                    abort();
                }
                break;
            case 'r':
                errno = 0;
                rowCount = strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"rowCount conv failed\n");
                    abort();
                }
                break;
            case 'c':
                errno = 0;
                columnCount = strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"columnCount conv failed\n");
                    abort();
                }
                break;
            case 's':
                errno = 0;
                randomSeed = strtol(optarg,NULL,10);
                if(errno) {
                    fprintf(stderr,"rowCount conv failed\n");
                    abort();
                }
                break;
            default:
                fprintf(stderr,"Usage: %s -a <agent-count> -t <time-step-count> -r <row-count> -c <column-count> -s <random-seed>\n",
                        argv[0]);
                exit(EXIT_FAILURE);
        }
    }

    if(timeStepCount < 0
            || agentCount < 0
            || columnCount < 0
            || rowCount < 0) {
        fprintf(stderr, "Error: User-supplied values must be positive.\n");
        abort();
    }

    if (agentCount > rowCount * columnCount) {
        fprintf(stderr, "Error: Too many agents for the given environment size.\n");
        abort();
    }

    srandom(randomSeed);

    /* Begin algorithm (see README) */

    /* 1: set grid size */
    cellCount = rowCount * columnCount;

    cellList  = (cell_t*)malloc(cellCount * sizeof(cell_t));
    agentList = (agent_t*)malloc(agentCount * sizeof(agent_t));

#pragma omp parallel for
    /* 2: start at cell 0 */
    for (cellIndex = 0;

            /* 3: is the cell outside the grid? */
            cellIndex < cellCount;

            /* 5: go to the next cell */
            cellIndex++) {

        /* 4: empty the cell */
        cellList[cellIndex].isEmpty = true;
        cellList[cellIndex].willBeEmpty = true;
    }

    /* 6: start at agent 0 */
    for (agentIndex = 0;

            /* 7: is the agent outside the list? */
            agentIndex < agentCount;

            /* 12: go to the next agent */
            agentIndex++) {

        /* 8: pick a random cell */
        cellIndex = cellCount * random() / RAND_MAX;

        /* 9: is the cell empty? */
        while (!cellList[cellIndex].isEmpty) {

            /* 10: try the next cell */
            if (cellIndex == cellCount - 1) {
                cellIndex = 0;
            } else {
                cellIndex++;
            }
        }

        /* 11: place the agent in the cell */
        cellList[cellIndex].isEmpty = false;
        cellList[cellIndex].willBeEmpty = false;
        agentList[agentIndex].cellIndex = cellIndex;
    }

    /* 13: start at time step 0 */
    for (timeStepIndex = 0;

            /* 14: have we finished? */
            timeStepIndex < timeStepCount;

            /* 40: move on to the next time step */
            timeStepIndex++) {

        /* 15: start at cell 0 */
        for (cellIndex = 0;

                /* 16: is the cell outside the grid? */
                cellIndex < cellCount;

                /* 19: go to the next cell */
                cellIndex++) {

            /* 17: print the cell */
            if (cellList[cellIndex].isEmpty) {
                printf("_");
            } else {
                printf("o");
            }

            /* 18: print newlines in between rows */
            if (cellIndex % columnCount == columnCount - 1) {
                printf("\n");
            }
        }

        /* 20: print a newline at the end */
        printf("\n");

#pragma omp parallel for private(cellIndex, direction, originalDirection, newCellIndex)
        /* 21: start at agent 0 */
        for (agentIndex = 0;

                /* 22: is the agent outside the list? */
                agentIndex < agentCount;

                /* 35: move on to the next agent */
                agentIndex++) {

            /* 23: set cell index */
            cellIndex = agentList[agentIndex].cellIndex;

            /* 24: pick a random number */
            direction = 4 * random() / RAND_MAX;
            originalDirection = direction;

            /* 25: set new cell index */
            if (direction == 0) {
                newCellIndex = cellIndex - 1;
            } else if (direction == 1) {
                newCellIndex = cellIndex + 1;
            } else if (direction == 2) {
                newCellIndex = cellIndex - columnCount;
            } else if (direction == 3) {
                newCellIndex = cellIndex + columnCount;
            } else {
                newCellIndex = cellIndex;
            }

            do {
                if (
                        /* 26: is left neighbor on the grid? */
                        (direction == 0 && cellIndex % columnCount == 0) ||

                        /* 27: is right neighbor on the grid? */
                        (direction == 1 &&
                         cellIndex % columnCount == columnCount - 1) ||

                        /* 28: is top neighbor on the grid? */
                        (direction == 2 &&
                         cellIndex < columnCount) ||

                        /* 29: is bottom neighbor on the grid? */
                        (direction == 3 &&
                         cellIndex >= (rowCount-1) * columnCount) ||

                        /* 30: will neighbor already be empty? */
                        (!cellList[newCellIndex].isEmpty ||
                         !cellList[newCellIndex].willBeEmpty)
                   ) {

                    /* 31: try the next neighbor */
                    if (direction == 3) {
                        direction = 0;
                    } else {
                        direction++;
                    }

                    /* 32: have we already tried this neighbor? */
                    if (direction == originalDirection) {
                        direction = 4;
                    }

                    /* 33: set new cell index */
                    if (direction == 0) {
                        newCellIndex = cellIndex - 1;
                    } else if (direction == 1) {
                        newCellIndex = cellIndex + 1;
                    } else if (direction == 2) {
                        newCellIndex = cellIndex - columnCount;
                    } else if (direction == 3) {
                        newCellIndex = cellIndex + columnCount;
                    } else {
                        newCellIndex = cellIndex;
                    }

                    /* 32: have we already tried this neighbor? */
                } else {
                    direction = 4;
                }
            } while (direction != 4);

            /* 34: update agent and cells */
            cellList[cellIndex].willBeEmpty = true;
            cellList[newCellIndex].willBeEmpty = false;
            agentList[agentIndex].cellIndex = newCellIndex;
        }

        /* 36: start at cell 0 */
        for (cellIndex = 0;

                /* 37: is the cell off the grid? */
                cellIndex < cellCount;

                /* 39: move on to the next cell */
                cellIndex++) {

            /* 38: advance the cell for the next time step */
            cellList[cellIndex].isEmpty = cellList[cellIndex].willBeEmpty;
        }
    }

    free(agentList);
    free(cellList);

    return 0;
}
